#ifndef bqt_dyn_h
#define bqt_dyn_h

#include "autoptr"
#include <cmath>
#include <iostream>
#include <list>

#define DYNDEBUG 1

/* Expression identifiers. Used for optimizing. */
typedef enum dyntype
{
    DYNTYPE_VALUE,
    DYNTYPE_PLUSMINUS,
    DYNTYPE_TIMESDIV,
    DYNTYPE_MOD,
    DYNTYPE_FUNC,
    DYNTYPE_DELETED
} dyntype;

#if DYNDEBUG > 0
static const void *addrmap[512]; static unsigned addrpos=0;
struct dynaddrname { unsigned ind; };
static dynaddrname AddrName(const void *p)
{
    if(p==NULL) { dynaddrname tmp; tmp.ind = 0; return tmp; }
    unsigned a;
    for(a=0; a<addrpos; ++a)
        if(p == addrmap[a])
            goto Done;
    addrmap[a = addrpos++] = p;
Done:
    dynaddrname tmp;
    tmp.ind = 0x111 + a * 4;
    if((long)p > 0)tmp.ind += 0x800;
    return tmp;
}
std::ostream &operator<< (std::ostream &dest, dynaddrname t)
{
    if(t.ind == 0)dest << "000";else
    dest << "0123456789ABCDEF"[(t.ind >> 8) & 15]
         << "0123456789ABCDEF"[(t.ind >> 4) & 15]
         << "0123456789ABCDEF"[(t.ind) & 15];
    return dest;
}
#endif

/* Throwable type: error code. */
class dynerror
{
public:
    enum type { getunsetval=1, getunsetexp, virtualcall, recursive, free_error };
private:
    type typi;
    const char *freemsg;
public:
    explicit dynerror(type t) : typi(t) { }
    explicit dynerror(const char *msg) : typi(free_error), freemsg(msg) { }
    operator const char * () const
    {
        return typi==getunsetval ? "Tried to get a value of UNSET variable"
             : typi==getunsetexp ? "Tried to get expression of UNSET variable"
             : typi==virtualcall ? "Pure virtual function called!"
             : typi==recursive ? "Recursive dynamic reference"
             : typi==free_error ? freemsg
             : "Failing dyn error";
    }
};

std::ostream &operator<< (std::ostream &dest, dyntype t)
{
    switch(t)
    {
        case DYNTYPE_VALUE: dest << "VALUE"; break;
        case DYNTYPE_PLUSMINUS: dest << "PLUSMINUS"; break;
        case DYNTYPE_TIMESDIV: dest << "TIMESDIV"; break;
        case DYNTYPE_MOD: dest << "MOD"; break;
        case DYNTYPE_FUNC: dest << "FUNC"; break;
        default: dest << '?' << (int)t << '?';
    }
    return dest;
}

/* Forward declaration (required by dyni::depends()) */
template<typename T>
class dyn;

/* Forward declaration (required by dyn::operator=()) */
template<typename T>
class dynvalue;








/* This is the base type for stored expressions.
 *
 */
template<typename T>
class dyni : public ptrable
{
public:
    dyni()
    {
#if DYNDEBUG > 0
        std::cout << "dyni born \33[32m[" << AddrName(this) << "]\33[37m\n";
#endif
    }
    virtual ~dyni()
    {
#if DYNDEBUG > 0
        std::cout << "dyni dying \33[32m[" << AddrName(this) << "]\33[37m\n";
        // debug(); std::cout << '"' << std::endl;
        //^ Can not do! The object has almost been deleted already!
#endif
    }
    virtual const T getvalue() const = 0;
    virtual void debug() const = 0; //{ throw dynerror(dynerror::virtualcall); }
    virtual const dyntype gettype() const = 0;
    virtual bool depends(dyn<T> &d) const = 0;
    virtual void undepend(dyn<T> &d) = 0;
};





/* This is the dynamic expression class.
 * It contains a pointer to the actual expression, which
 * may be shared by multiple dyn-instances. The expression
 * contains a reference counter, and is automatically
 * destructed when the last dyn pointing to it dies.
 */
typedef enum { t_normal=0, t_deleted, t_tempcopy } dynstatetype;
template<typename T>
class dyn: public ptrable
{
private:
    autoptr<dyni<T> > v;
    dynstatetype type;
public:
    explicit dyn(dyni<T> *q);
    dyn(const dyn<T> &);
    dyn(const T &);

    dyn(): type(t_normal) { }
    virtual ~dyn()
    {
#if DYNDEBUG > 0
#if DYNDEBUG <= 3
        if(type != t_tempcopy)
#endif
        { 
        std::cout << "dyn dying: " << std::flush;
        debug(); std::cout << std::endl; }
        v = NULL;
#if DYNDEBUG <= 3
        if(type != t_tempcopy) 
#endif
        {
        std::cout << "dyn died: " << std::flush;
        debugaddr(); std::cout << std::endl; }
#endif
        type = t_deleted;
    }
    
    const T getvalue() const
    {
        if(!v) throw dynerror(dynerror::getunsetval);
        return v->getvalue();
    }
#if DYNDEBUG > 0
    void debugaddr() const
    {
        std::cout << "\33[31m[\33[33m" << get_autoptr_refnum()
                  << "\33[31m:" << AddrName(this);
        if(v) { std::cout
                  << "\33[32m/\33[36m"
                  << v->get_autoptr_refnum()
                  << "\33[32m:"
                  << AddrName(v) << "\33[31m]";
        } else std::cout << "\33[34m]";
        std::cout << "\33[37m";
    }
#endif
    void debug() const
    {
        debugaddr();
        if(type == t_deleted) { std::cout << "(DELETED)"; return; }
        if(!v) { std::cout << "(NOT SET)"; return; }
        v->debug();
    }
    dyn &operator= (const dyn<T> &p);
    dyn &operator= (const T &value); // constant
    const dyn &operator= (dyni<T> *q);
    
    /* typecast operator */
    //operator T () const { return getvalue(); }
    
    const dyni<T> *const getexpr() const
    {
        if(!v) throw dynerror(dynerror::getunsetexp);
        return v;
    }
    bool depends(dyn<T> &d) const
    {
        if(&d == this)return true;
        if(!v)throw dynerror(dynerror::getunsetval);
        return v->depends(d);
    }
    void undepend(dyn<T> &d) const { v->undepend(d); }
};

template<typename T>
dyn<T>::dyn(dyni<T> *p) : v(p), type(t_tempcopy)
{
#if DYNDEBUG > 3
    if(p){std::cout << "dyn created from dyni: " << std::flush;
    debug(); std::cout << " (" << v->gettype() << ')' << std::endl;}
#endif
}

template<typename T> dyn<T>::dyn(const dyn<T> &c) :ptrable(), v(c.v),type(t_normal) { }
template<typename T> dyn<T>::dyn(const T &c) : v(new dynvalue<T>(c)),type(t_normal) { }

/* Reassignments */
template<typename T> dyn<T> &dyn<T>::operator= (const dyn<T> &p)
{
    if(this == &p)
    {
#if DYNDEBUG > 0
        std::cout << "Skipping redundant self-assignment\n";
#endif
        return *this; /* redundant self-assignment ( a=a; ) */
    }
    if(p.depends(*this))
    {
        /* Setting this to something that depends on self. */
        /* This is a special case. */
        
#if DYNDEBUG > -1
        std::cout << "Undepending1\n";
#endif
        p.v->undepend(*this);
#if DYNDEBUG > -1
        std::cout << "Done undepending\n";
#endif
        
        //throw(dynerror(dynerror::recursive));
    }
    v = p.v;
#if DYNDEBUG > 0
    if(v){std::cout << "dyn received from dyni: " << std::flush;
    debug(); std::cout << " (" << v->gettype() << ')' << std::endl;}
#endif
    return *this;
}
template<typename T> const dyn<T> &dyn<T>::operator= (dyni<T> *p)
{
    /* Called in return (Dest = p) -style clauses. */
    if(p->depends(*this))
    {
        /* Setting this to something that depends on self. */
        /* This is a special case. */

#if DYNDEBUG > -1
        std::cout << "Undepending2\n";
#endif
        p->undepend(*this);
#if DYNDEBUG > -1
        std::cout << "Done undepending\n";
#endif
        
        //throw(dynerror(dynerror::recursive));
    }
    v = p;
    return *this;
}

template<typename T> dyn<T> &dyn<T>::operator= (const T &value)
{
    v = new dynvalue<T> (value);
    return *this;
}





/* Expression type: Constant */
template<typename T>
class dynvalue : public dyni<T>
{
    T v;
public:
    explicit dynvalue(const T &c) : v(c) { }
    virtual const T getvalue() const
    {
        return v;
    }
    virtual void debug() const { std::cout << v << std::flush; }
    virtual const dyntype gettype() const { return DYNTYPE_VALUE; }
    virtual bool depends(dyn<T> &) const { return false; }
    virtual void undepend(dyn<T> &) { }
};













/* This is a base class containing a garbage colletion.
 * All elements added are guaranteed to stay
 * alive at least until this class dies. */
template<typename T>
class dyntmpkeep
{
private:
    typedef autoptr<dyn<T> > tmptype;
    std::list<tmptype> tmplist;
public:
    void addtmp(dyn<T> *p)
    {
#if DYNDEBUG > 0
        std::cout << "Added temporary: ";
        p->debug();
        std::cout << std::endl;
#endif
        /* Add p to the automatic temporary variable deletion list 
         * (They will be deleted when the destructor is called, due to autoptr.)
         */
        tmplist.push_back(p);
    }
    void deltmp(dyn<T> *p)
    {
        typename std::list<tmptype>::iterator i;
        for(i=tmplist.begin(); i!=tmplist.end(); ++i)
            if(*i == p)
            {
#if DYNDEBUG > 0
                std::cout << "Deleting temporary: ";
                p->debug();
                std::cout << std::endl;
#endif
                tmplist.erase(i);
                return;
            }
    }
    dyntmpkeep()
    {
#if DYNDEBUG > 0
        std::cout << "Created empty temporary store\n";
#endif
    }
    virtual ~dyntmpkeep()
    {
#if DYNDEBUG > 0
        if(!tmplist.size())
        {
            std::cout << "Deleting empty temporary store\n";
        }
        else
        {
            std::cout << "Deleting temporaries\n";
            typename std::list<tmptype>::const_iterator i;
            for(i=tmplist.begin(); i!=tmplist.end(); ++i)
            {
                std::cout << ": ";
                (*i)->debug();
                std::cout << std::endl;
            }
            std::cout << "[end of list]\n";
        }
#endif
    }
};





template<typename T>
class dynsigncoll : public dyni<T>, private dyntmpkeep<T>
{
private:

    /* This class is a contained class, not an anchestor for anything! */
    class dynsignpair : public ptrable
    {
    private:
        dynsignpair &operator= (const dynsignpair &)
        {
            throw dynerror("dynsignpair::operator=(const dynsignpair &) called");
        }
#if DYNDEBUG > 1
        void debuggi(const char *s)
        {
            std::cout << s << "ed dynsignpair: " << (nega?"-/ " : "+* ");
            expr->debug();
            std::cout << std::endl;
        }
#endif
    public:
        bool nega;
        
        /* Don't make this autoptr! This might be the only pointer existing
         * to a local variable, which still exists in a stack frame.
         */
        dyn<T> *expr;
        //autoptr<dyn<T> > expr;
    public:
        dynsignpair(bool neg, dyn<T> &exp) : nega(neg), expr(&exp)
        {
#if DYNDEBUG > 1
            debuggi("Creat");
#endif
        }
        /* Virtual is not needed, as this isn't a base class,
         * and deletion is done by autoptr, using dynsignpair, not ptrable.
         */
        ~dynsignpair()
        {
            /* Don't deallocate expr here. See expr def. */
#if DYNDEBUG > 1
            debuggi("Kill");
#endif
        }
        
        bool undepend(dyn<T> &exp)
        {
            expr->undepend(exp);
            if(expr == &exp)
            {
                const dyn<T> &f = *expr;
#if DYNDEBUG > -1
                std::cout << "Replacing in \33[35m[" << AddrName(this) << "]\33[37m: "; expr->debug(); 
#endif
                expr = new dyn<T> (f);
#if DYNDEBUG > -1
                std::cout << "\n               with "; expr->debug();
                std::cout << /*" (" << AddrName(expr) << ')' <<*/ std::endl;
#endif
                return true;
            }
            return false;
        }
    };

    typedef autoptr<dynsignpair> elemtype;
    typedef std::list<elemtype> elemlist;
    typedef typename elemlist::const_iterator const_elemit;

    elemlist elems;

    char sgneg, sgpos;

public:
    dynsigncoll(char n, char p) : sgneg(n), sgpos(p)
    {
#if DYNDEBUG > 1
        std::cout << "Created dynsigncoll('" << n << "', '" << p << "')\n";
#endif
    }
    virtual ~dynsigncoll()
    {
#if DYNDEBUG > 1
        std::cout << "Killing dynsigncoll \"" << std::flush;
        debug();
        std::cout << '"' << std::endl;
#endif
    }
    virtual void debug() const;
    void add(dyn<T> &elem, bool nega);
    void add(dyn<T> *elem, bool nega);
    virtual bool depends(dyn<T> &d) const;
    virtual void undepend(dyn<T> &);

    virtual const T getvalue() const;
    virtual const T emptyvalue() const = 0;
    virtual void evaluate(T &seed, bool nega, const T &value) const = 0;
    
    dynsigncoll *Optimize()
    {
        /* TODO: If self is a multiplication list, replace self with adding
         * list including elements multiplied together. Thus, this function
         * should call "delete this".
         */
        return this;
    }
};

template<typename T>
const T dynsigncoll<T>::getvalue() const
{
    T result = emptyvalue();
    for(const_elemit i = elems.begin(); i != elems.end(); ++i)
        evaluate(result, (*i)->nega, (*i)->expr->getvalue());
    return result;
}

template<typename T>
void dynsigncoll<T>::debug() const
{
    if(!elems.size()) { std::cout << "[EMPTY]"; return; }
    if(elems.size() > 1)std::cout << '(';
    const_elemit i = elems.begin();
    (*i)->expr->debug();
    while(++i != elems.end())
    {
        std::cout << (char)((*i)->nega ? sgneg : sgpos);
        (*i)->expr->debug();
    }
    if(elems.size() > 1)std::cout << ')';
}

template<typename T>
void dynsigncoll<T>::add(dyn<T> *elem, bool nega)
{
    addtmp(elem);
    add(*elem, nega);
}

template<typename T>
void dynsigncoll<T>::add(dyn<T> &elem, bool nega)
{
    const dyni<T> *expr = elem.getexpr();
#if DYNDEBUG > 2
    std::cout << "ADD " << gettype() << ": "; debug(); std::cout << " <- " << expr->gettype() << ": "; expr->debug(); std::cout << std::endl;
#endif
    
    /* If adding a collection to a collection. */
    if(expr->gettype() == gettype())
    { 
        /* For unknown reason, dynamic_cast causes segmentation fault.
         * static_cast does not, and works as supposed.
         */
        const dynsigncoll<T> *expc = static_cast<const dynsigncoll<T> *> (expr);
        const elemlist e = expc->elems; /* May be self-reference. Thus copy. */
        for(const_elemit i=e.begin(); i!=e.end(); ++i)
            elems.push_back(new dynsignpair(nega ^ (*i)->nega, *(*i)->expr));
    }
    else
    {
#if 0 /* EXPANSION -- TODO */
        /* If adding a sum to a multiplication list */
        if(gettype() == DYNTYPE_TIMESDIV
        && expr->gettype() == DYNTYPE_PLUSMINUS)
        {
            
        }
#endif
        /* Add the element as requested. */
        elems.push_back(new dynsignpair(nega, elem));
    }
    //std::cout << "DONE" << std::endl;
}

template<typename T>
bool dynsigncoll<T>::depends(dyn<T> &d) const
{
    const_elemit i;
    for(i=elems.begin(); i!=elems.end(); ++i)
        if((*i)->expr->depends(d))return true;
    return false;
}

template<typename T>
void dynsigncoll<T>::undepend(dyn<T> &d)
{
    const_elemit i;
    for(i=elems.begin(); i!=elems.end(); ++i)
    {
        dynsignpair *p = (*i);
        dyn<T> *oldexpr = p->expr;
        if(p->undepend(d))
        {
            deltmp(oldexpr);
            /* dynsignpair doesn't take care of deletion. */
            addtmp(p->expr);
        }
    }
}




template<typename T>
class dynplusminus : public dynsigncoll<T>
{
public:
    dynplusminus() : dynsigncoll<T> ('-', '+') { }
    virtual const dyntype gettype() const { return DYNTYPE_PLUSMINUS; }

    virtual const T emptyvalue() const { return 0; }
    virtual void evaluate(T &seed, bool nega, const T &value) const
    {
        if(nega) seed -= value;
        else seed += value;
    }
};

template<typename T>
class dyntimesdiv : public dynsigncoll<T>
{
public:
    dyntimesdiv() : dynsigncoll<T> ('/', '*') { }
    virtual const dyntype gettype() const { return DYNTYPE_TIMESDIV; }

    virtual const T emptyvalue() const { return 1; }
    virtual void evaluate(T &seed, bool nega, const T &value) const
    {
        if(nega) seed /= value;
        else seed *= value;
    }
};

/* Plenty of operators to be defined. Doing them with macroes. */

#define dyn2opdef2(op, nega, obj, p1,pc1, p2, pc2) \
  template<typename T> \
  const dyn<T> operator op (p1 &L, p2 &R) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(pc1 (L), false); \
    p->add(pc2 (R), nega); \
    return dyn<T> (p->Optimize()); \
  }
#define dyn2opdef(op, nega, obj) \
  dyn2opdef2(op, nega, obj, const dyn<T>, new dyn<T>, const dyn<T>, new dyn<T>) \
  dyn2opdef2(op, nega, obj,       dyn<T>,           , const dyn<T>, new dyn<T>) \
  dyn2opdef2(op, nega, obj, const dyn<T>, new dyn<T>,       dyn<T>,           ) \
  dyn2opdef2(op, nega, obj,       dyn<T>,           ,       dyn<T>,           ) \
  dyn2opdef2(op, nega, obj, const dyn<T>, new dyn<T>, const     T , new dyn<T>) \
  dyn2opdef2(op, nega, obj,       dyn<T>,           , const     T , new dyn<T>) \
  dyn2opdef2(op, nega, obj, const     T , new dyn<T>, const dyn<T>, new dyn<T>) \
  dyn2opdef2(op, nega, obj, const     T , new dyn<T>,       dyn<T>,           ) \

dyn2opdef(+, false, dynplusminus)
dyn2opdef(-, true,  dynplusminus)
dyn2opdef(*, false, dyntimesdiv)
dyn2opdef(/, true,  dyntimesdiv)

#undef dyn2opdef
#undef dyn2opdef2

#define dyn2opdef2(opa, nega, obj, p2, pc2) \
  template<typename T> \
  const dyn<T> operator opa (dyn<T> &dest, p2 &R) \
  { \
    dyn<T> *tmp = new dyn<T> (dest); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(pc2 (R), nega); \
    return (dest = p->Optimize()); \
  }
#define dyn2opdef(opa, nega, obj) \
  dyn2opdef2(opa, nega, obj, const dyn<T>, new dyn<T>) \
  dyn2opdef2(opa, nega, obj,       dyn<T>,           ) \
  dyn2opdef2(opa, nega, obj, const     T , new dyn<T>)

dyn2opdef(+=, false, dynplusminus)
dyn2opdef(-=, true,  dynplusminus)
dyn2opdef(*=, false, dyntimesdiv)
dyn2opdef(/=, true,  dyntimesdiv)

#undef dyn2opdef
#undef dyn2opdef2


#if 1 /* mod_and_fun */



/* Expression base type: Modulo */
template<typename T>
class dynmod : public dyni<T>, public dyntmpkeep<T>
{
private:
    autoptr<dyn<T> > L, R;
public:
    dynmod(const dyn<T> &l, const dyn<T> &r) : L(new dyn<T> (l)), R(new dyn<T> (r)) { addtmp(L); addtmp(R); }
    dynmod(const dyn<T> &l, dyn<T> &r) : L(new dyn<T> (l)), R(&r) { addtmp(L); }
    dynmod(dyn<T> &l, const dyn<T> &r) : L(&l), R(new dyn<T> (r)) { addtmp(R); }
    dynmod(dyn<T> &l, dyn<T> &r) : L(&l), R(&r) {}
    virtual const T getvalue() const { return L->getvalue() % R->getvalue(); }
    virtual void debug() const { std::cout << '{'; L->debug();
                                 std::cout << '%'; R->debug();
                                 std::cout << '}'; }
    virtual const dyntype gettype() const { return DYNTYPE_MOD; }
    virtual bool depends(dyn<T> &d) const { return L->depends(d) || R->depends(d); }
    virtual void undepend(dyn<T> &);
};
template<typename T> const dyn<T> operator % (const dyn<T> &l, const dyn<T> &r) { return dyn<T> (new dynmod<T> (l, r)); }
template<typename T> const dyn<T> operator % (      dyn<T> &l, const dyn<T> &r) { return dyn<T> (new dynmod<T> (l, r)); }
template<typename T> const dyn<T> operator % (const dyn<T> &l,       dyn<T> &r) { return dyn<T> (new dynmod<T> (l, r)); }
template<typename T> const dyn<T> operator % (      dyn<T> &l,       dyn<T> &r) { return dyn<T> (new dynmod<T> (l, r)); }
template<typename T>
const dyn<T> &operator %= (dyn<T> &dest, const dyn<T> &r)
{
    dyn<T> *tmp = new dyn<T> (dest);
    dynmod<T> *p = new dynmod<T> (*tmp, r);
    p->addtmp(tmp);
    return (dest = p);
}
template<typename T>
const dyn<T> &operator %= (dyn<T> &dest, dyn<T> &r)
{
    dyn<T> *tmp = new dyn<T> (dest);
    dynmod<T> *p = new dynmod<T> (*tmp, r);
    p->addtmp(tmp);
    return (dest = p);
}
template<typename T>
void dynmod<T>::undepend(dyn<T> &d)
{
    L->undepend(d); R->undepend(d);
    if(L->depends(d)) { const dyn<T> &L2 = *L; addtmp(L = new dyn<T> (L2)); }
    if(R->depends(d)) { const dyn<T> &R2 = *R; addtmp(R = new dyn<T> (R2)); }
}






template<typename T>
class dynfun : public dyni<T>, dyntmpkeep<T>
{
private:
    autoptr<dyn<T> > E;
    T (*const F) (T);
    const char *funame;
public:
    dynfun(const dyn<T> &exp, T (*const fun) (T), const char *f);
    dynfun(dyn<T> &exp, T (*const fun) (T), const char *f);
    virtual ~dynfun();
    virtual const T getvalue() const { return F(E->getvalue()); }
    virtual void debug() const;
    virtual const dyntype gettype() const { return DYNTYPE_FUNC; }
    virtual bool depends(dyn<T> &d) const { return E->depends(d); }
    virtual void undepend(dyn<T> &);
};
template<typename T>
dynfun<T>::dynfun(const dyn<T> &exp, T (*const fun) (T),  const char *f)
             : E(new dyn<T> (exp)), F(fun), funame(f)
{
    addtmp(E);
#if DYNDEBUG > 0
    std::cout << "Created dynfun(const "; E->debug();
    std::cout << ", \"" << f << "\")\n";
#endif
}
template<typename T>
dynfun<T>::dynfun(dyn<T> &exp, T (*const fun) (T), const char *f)
             : E(&exp), F(fun), funame(f)
{
#if DYNDEBUG > 0
    std::cout << "Created dynfun("; E->debug();
    std::cout << ", \"" << f << "\")\n";
#endif
}
template<typename T>
dynfun<T>::~dynfun()
{
#if DYNDEBUG > 0
    std::cout << "Killed dynfun: ";
    E->debug();
    std::cout << std::endl;
#endif
}

template<typename T>
void dynfun<T>::debug() const
{
    std::cout << funame << '('; E->debug();
    std::cout << ')';
}

template<typename T>
void dynfun<T>::undepend(dyn<T> &d)
{
    E->undepend(d);
    if(E->depends(d)) { const dyn<T> &E2 = *E; addtmp(E = new dyn<T> (E2)); }
}

#define dynfundef2(fu, fus) \
template<typename T> \
inline const dyn<T> fus(const dyn<T> &e) \
{ \
    return dyn<T> (new dynfun<T> (e, fu, #fus)); \
} \
template<typename T> \
inline const dyn<T> fus(dyn<T> &e) \
{ \
    return dyn<T> (new dynfun<T> (e, fu, #fus)); \
}

/* This logic is from /usr/include/math.h in glibc 2.2.4 */
#if defined __USE_MISC || defined __USE_ISOC99
# if (__STDC__ - 0 || __GNUC__ - 0) && !defined __NO_LONG_DOUBLE_MATH
#  define dynfundef(fu) \
  dynfundef2(std::fu,   fu) \
  dynfundef2(fu##f,     fu##f) \
  dynfundef2(fu##l,     fu##l)
# else
#  define dynfundef(fu) \
  dynfundef2(std::fu,   fu) \
  dynfundef2(fu##f,     fu##f)
# endif
#else
# define dynfundef(fu) \
  dynfundef2(std::fu,   fu)
#endif

dynfundef(acos)
dynfundef(asin)
dynfundef(atan)
dynfundef(ceil)
dynfundef(cos)
dynfundef(cosh)
dynfundef(exp)
#if defined __USE_MISC || defined __USE_XOPEN_EXTENDED || defined __USE_ISOC99
 dynfundef(expm1)
 dynfundef(log1p)
 dynfundef(logb)
 dynfundef(cbrt)
# ifdef __USE_ISOC99
  dynfundef(exp2)
  dynfundef(log2)
  dynfundef(round)
# endif
# ifdef __USE_MISC
  dynfundef(significand)
# endif
#endif
#ifdef __USE_GNU
 dynfundef(sincos)
#endif
dynfundef(floor)
dynfundef(log)
dynfundef(log10)
dynfundef(sin)
dynfundef(sinh)
dynfundef(sqrt)
dynfundef(tan)
dynfundef(tanh)

#undef dynfundef

#endif /* mod_and_fun */

#endif
